#ifndef SKETCHPAINTER_H
#define SKETCHPAINTER_H

#include <qrect.h>
#include <qcstring.h>
#include <qpen.h>
#include <qobject.h>
#include <qbitarray.h>

#include "constants.h"

class QColor;
class QCString;
class QPointArray;

////////////////////////////////////////////////////////////
// PenShape

class PenShape {
  double half_width;
  QByteArray table;
  QByteArray alphas;

public:
  PenShape(void);

  void setHalfWidth(double);
  void setWidth(const QPen&);
  double halfWidth(void) {return half_width;};

  char width(uint);
  uchar alpha(uint);
};

////////////////////////////////////////////////////////////
// SketchPainter

class SketchPainter : public QObject {
  Q_OBJECT

protected:
  unsigned short int *buf;
  static uint32 *cloud_work;
  static QBitArray *cloud_on;
  static QBitArray *mask;
  int width, height, orientation;
  QPen pen;
  QColor paper_color;
  int brush_method;
  uint density, cloud_density;
  bool antialias;
  int composition_method;
  uint composition_alpha;
  bool showing;
  PenShape pen_shape;
  QPoint shift;
  bool _modified;

public:
  SketchPainter(int w = 0, int h = 0, int o = 0);
//  SketchPainter(const QSize &, int o = 0);
  ~SketchPainter() {
    delete buf;
  }

  void initialize(int, int, int = 0);

  static const uint pensize_tiny = 2;
  static const uint pensize_small = 6;

  void setSize(int, int, int = 0);
  QPen get_pen(void) {return pen;};
  int get_width(void) {return width;};
  int get_height(void) {return height;};
  
  void fill(const QColor &c);
  void fill(void);
  unsigned short int* frameBuffer(void);

//  void setCloudWork(unsigned long int*, QBitArray*);
  void setOrientation(int);

  QString infoString(void);
  
  void setPen(const QPen &);
  void setPenColor(const QColor &);
  void setPenMethod(int);
  int penMethod(void);
  void setPenDensity(uint);
  uint penDensity(void) {return density;};
  void setCloudDensity(uint);
  uint cloudDensity(void) {return cloud_density;};
  void setAntialias(bool);
  void setShowing(bool);
  bool getShowing(void) {return showing;}

  void clearMask(void);
  
  void drawPoint(int x, int y);
  void drawPoint(const QPoint &p);
  void drawLine(int x0, int y0, int x1, int y1);
  void drawLine(const QPoint &p0, const QPoint &p1);
  void drawPolyline(const QPointArray &pta);
  void drawWaterLine(int, int, int, int, uint = 255);
  void drawWaterLine(const QPoint&, const QPoint&, uint = 255);

  void mirrorHorizontal(void);
  void mirrorVertical(void);
  void rotateCW(void);
  void rotateCCW(void);
  
  void fillRect16(int, int, int, int, unsigned short int);
  void fillRect16(const QRect&, unsigned short int);
  void fillRect32(int, int, int, int, unsigned long int);
  void fillRect32(const QRect&, unsigned long int);

  void invertRect(int, int, int, int);
  void invertRect(const QRect&);
  
  QColor pickColor(int, int);

  QColor paperColor(void) {return paper_color;};
  void setPaperColor(QColor p) {paper_color = p;};
  int compositionAlpha(void) {return composition_alpha;}
  void setCompositionAlpha(int a) {composition_alpha = a;}
  int compositionMethod(void) {return composition_method;}
  void setCompositionMethod(int m) {composition_method = m;}
  
  void getQImageR0(QImage *, const QRect& = QRect());
  void getQImageR270(QImage *, const QRect& = QRect());
  QImage getQImage();
  void getQImage(QImage*, int);
  void getQImage(QImage*, const QRect& = QRect(), int = -1);
  void setQImage(const QImage&);
  bool load(const QString &, const char* = 0);
  bool save(const QString &, const char*);
  bool load_raw(const QString &);
  bool save_raw(const QString &);

  static unsigned short int pack_color(const QColor &c);
  static QColor unpack_color(unsigned short int);
  

protected:
  void draw_line(int, int, int, int, bool = true);
  void draw_line(const QPoint&, const QPoint&, bool = true);
  void draw_diamond_line(int, int, int, int, bool = true);
  void draw_diamond_line(const QPoint&, const QPoint &,
                         bool = true);
  void draw_diamond_line_w1(int, int, int, int);
  void draw_diamond_line_w2(int, int, int, int);
  void draw_water_line(int, int, int, int, uint = 255);
  void draw_water_line(const QPoint&, const QPoint&, uint = 255);
  void draw_water_line_w1(int, int, int, int, uint = 255);
//  void draw_point(int, int, float);
  void draw_point(int, int, unsigned int = 255);
  void draw_diamond(int, int, bool = true);
  void draw_diamond(const QPoint &, bool = true);
  void draw_circle(int, int, bool = true);
  void draw_circle(const QPoint &, bool = true);
  void draw_water_diamond(int, int, uint = 255);
  void draw_water_diamond(const QPoint &, uint = 255);
  void draw_water_circle(int, int, uint = 255);
  void draw_water_circle(const QPoint &, uint = 255);
  void draw_cloud_line(int, int, int, int, bool);
  void draw_cloud_short_line(int, int, int, int, bool);

  QPoint& copyWithShift(uint16*, QPoint &);

  void clear_cloud_work(int, int);
  void set_cloud_flag(int, int, int, int, int, int);
//  void set_cloud_blend(int, int, int, int, int);
  void disperse_rect(int, int, int, int);
  void disperse_h_rect(int, int, int, int);

  void clip_to_this(int *, int *, int = 0);
  void clip_to_this(int &, int &, int = 0);
  void swap_points(int *, int *, int *, int *);
  void swap_points(int &, int &, int &, int &);
  int get_index(int x, int y);
  QPoint get_point(int);
  //  inline unsigned long int blend(unsigned long int&,
  //                                 unsigned long int&,
  //                                 unsigned long int&, uint&,
  //                                 uint16, uint16, uint16);
  inline unsigned long int SketchPainter::blend(
                                                unsigned long int &pack, unsigned long int &rgb,
                                                unsigned long int &p_rgb, uint &alpha,
                                                uint16 pred, uint16 pgreen, uint16 pblue) {
    rgb = ((pack & 0xf800) << 9) |
      ((pack & 0x07c0) << 4) |
      (pack & 0x001f);
    rgb *= 32 - alpha;
    rgb += p_rgb;

    rgb = ((rgb & 0x3e000000) >> 14) |
      ((rgb & 0x000f8000) >> 9) |
      ((rgb & 0x000003e0) >> 5);

    if (pack == rgb) {
      if ((pack & 0xf800) < pred)
        rgb += 0x0800;
      else if ((pack & 0xf800) > pred)
        rgb -= 0x0800;

      if ((pack & 0x07c0) < pgreen)
        rgb += 0x0040;
      else if ((pack & 0x07c0) > pgreen)
        rgb -= 0x0040;

      if ((pack & 0x001f) < pblue)
        rgb ++;
      if ((pack & 0x001f) > pblue)
        rgb --;
    }
    
    return rgb;
  }
  inline uint32 SketchPainter::pick_dispersed_pixel(
                                                    uint cindex, uint32 opixel, bool edge) {
    if (edge || cloud_on->at(cindex))
      return cloud_work[cindex];
    else
      return opixel;
  }

  inline uint32 SketchPainter::unpack_color_uint32(uint16 p) {
    return ((p & 0xf100) << 8) |
      ((p & 0x07e0) << 6) |
      ((p & 0x001f) << 3);
  }



#include "painter.inline"
  
 public slots:
  void loadOverLayer(QString);
  void saveOverLayer(QString);

 private:
  QString overLayerName(QString);
};


#endif	// SKETCHPAINTER_H
